/*************************************************************************
 * Author: Brian Chan (birdchan@ucdavis.edu)	Date: Jan 14th, 2002
 * Supervisor: Alex Kozik	Dept: Vegetable and Crop Dept (UC Davis)
 *
 * This program reads in a list file containing a set of nodes. Then it
 *   reads in the edges from the matrix file and put corresponding 
 *   nodes into groups. The edges are acceptable if the bonding is 
 *   higher than the cutoff value (0<=x<=1) inputted by the user. 
 * Once all groups are formed (matrix file is done reading), the group
 *   infomation will be outputed. 
 *
 **********************************************************************/

#include <stdlib.h>
#include <stdio.h>
//#include <iostream.h>
#include <iostream>
//#include <fstream.h>
#include <fstream>
#include <string.h>
#include <algorithm>
#include <set>
#include <string>
#include <sstream>
#include "seqNameClass.h"
#include "groupUsedClass.h"
#include "groupsClass.h"
#include "matrixClass.h"
#include "seqRelationClass.h"

#define MAXLEN 80     // for getline 

using std::ifstream;
using std::ofstream;
using std::cin;
using std::cout;
using std::endl;
using std::ios;
using std::ostringstream;



	// Read in filenames and cutoff value
void init(ifstream& list_fp, ifstream& matrix_fp, float* cutoff_value);
	// get seq name from the list file
void getSeqName(ifstream& list_fp, long* numSeq);
	// form the groups
void setGroups(ifstream& matrix_fp, float cutoff_value, long numSeq);
	// retrieve the internal index of a seq name
long  getSeqIndex(long numSeq, string seq);
	// print out groups
void printGroups(long numSeq);
	// print out group details
void printGroupsDetails(long numSeq);
	// get available group index
long getGroupIndex();
	// output groups to files
long outputGroups(int step, long numSeq, bool *isXBridge, bool *isYBridge);
	// print the usage of the parameters, and quit
void printUsage();
	// analyze the subgroups
long checkMatrixGroups(ifstream& matrix_fp, float cutoff_value,
	 long numSeq, string skipSeq);

// ---------------------------------------------------------------------

int g_verbose = 0; 	  // display current edges    
int g_debug_01 = 0;       // show each step during grouping (in matrix file)
int g_debug_02 = 0;       // if DEBUG_01 enable, it will enable the pause
int g_debug_03 = 0;       // show each step during assigning new group to
			  // those alone (not in the matrix file)
long maxseq = 200;
string listFileName;

seqNameClass seqName;
groupUsedClass groupUsed;
groupsClass groups;
matrixClass matrix;
seqRelationClass seqRelation;

// ---------------------------------------------------------------------

// The main program
int main(int argc, char* argv[]){

		// set global vars
	if( argc > 1 ){
		for(int i=2; i<=argc; i++){
			if(argv[i-1][0] == '-'){
				switch(argv[i-1][1]){
					case 'v': g_verbose = 1;
						  break;
					case 'd': switch(argv[i-1][2]){
							  case '1':
							    g_debug_01 = 1;
							    break;
							  case '2':
							    g_debug_02 = 1;
							    break;
							  case '3':
							    g_debug_03 = 1;
							    break;
						  }
						  break;
					default: printUsage();
				}
			}else{
				printUsage();
			}
		}
	}
	
	ifstream list_fp;              // list file handler
	ifstream matrix_fp;            // matrix file handler
	float cutoff_value = 0.0;      // cut-off value
	long numSeq = 0;               // # of seq ( should < MAXSEQ )
	long numGroup = 0;             // # of groups

	init(list_fp, matrix_fp, &cutoff_value);    // get related info
	string find_bridges;
	cout << "Do you like to find bridges in subgroups? (y/n): ";
	cout.flush();
	cin >> find_bridges;

	cout << "Reading sequence names...";
	cout.flush();
	getSeqName(list_fp, &numSeq);      // get seq names
	cout << "  done." << endl;
		
		// form the groups
	cout << "Reading the relations...";
	cout.flush();
	setGroups(matrix_fp, cutoff_value, numSeq);
	cout << "  done." << endl;
		
		// close file pointers
	list_fp.close();
	matrix_fp.close();

		// print out all the groups
	if(g_verbose){
		cout << "\n\nHere are the final groups:" << endl;
		cout << "----------------------------------\n" << endl;
		printGroups(numSeq);
	}

		// output subgroups to files
	cout << "Writing to the subgroup files...";
	cout.flush();
		// 1 = to output subgroups and matrices
	numGroup = outputGroups(1, numSeq, NULL, NULL);
	cout << "  done." << endl;

		// Find bridges
	long count = 0;
	bool isXBridge[maxseq];
	bool isYBridge[maxseq];
	for(long i=0; i<maxseq; i++){
		isXBridge[i] = false;
		isYBridge[i] = false;
	}
	// find bridges -----------------------------------
	if(find_bridges=="y"){
		// find X bridges ------------------------------
		// X bridge is a bridge which if not present, will
		//   create two or more subgroups
		for(long i=0; i<maxseq; i++){   // for each group
			long numOfSeq = groups[i].size();
			if(groups[i].size()<=0) continue;
			count++;
			set<long>::const_iterator it;
			it = groups[i].begin();
			for(long j=0; j<numOfSeq; j++){  // for each seq
				long temp = count;
				string prefix = "submatrix_";
				string suffix = "";
				while(temp!=0){
					string c;
					switch(temp % 10){
						case 0: c = "0"; break;
						case 1: c = "1"; break;
						case 2: c = "2"; break;
						case 3: c = "3"; break;
						case 4: c = "4"; break;
						case 5: c = "5"; break;
						case 6: c = "6"; break;
						case 7: c = "7"; break;
						case 8: c = "8"; break;
						case 9: c = "9"; break;
						default: break;
					}
					temp = temp / 10;  // shift it by one digit
					suffix = c + suffix;
				}
				if(g_verbose){
					cout << "checking... seq " << seqName[*it]
					     << " in " << prefix+suffix << endl;
				}
				string filename = prefix + suffix;
				ifstream inf(filename.c_str(), ios::in);
				long num;
				num = checkMatrixGroups(inf, cutoff_value,
				         numSeq, seqName[*it]);
				if(num>1){
					isXBridge[*it] = true;
					if(g_verbose){
						cout << seqName[*it] << " is a X bridge" << ", isXBridge[" << *it << "] is true " << endl;
					}
				}
				inf.close();
				it++;
			}
		}
		// find Y bridges ------------------------------
		// Y bridge is a bridge which if is absent, will
		//   create one-element subgroup(s)
		for(long i=0; i<maxseq; i++){   // for each group
			long numOfEdges = seqName.loadNeighbor(i).size();
			if(numOfEdges==1){    // we are only interested in this
				set<long>::const_iterator it;
				it = (seqName.loadNeighbor(i)).begin();
				long noe = seqName.loadNeighbor(*it).size();
				if (noe > 1){
					isYBridge[*it] = true;
					if(g_verbose){
						cout << seqName[*it] << " is a Y bridge" << ", isYBridge[" << *it << "] is true " << endl;
					}
				}
			}
		}
	}
	// done find bridges -----------------------------------------
	cout << "----------------------------------" << endl;
	for(long i=0; i<maxseq; i++){
		if(isXBridge[i] && isYBridge[i]){
			cout << seqName[i] << " is a Z bridge" << endl;
		}else
		if(isXBridge[i]){
			cout << seqName[i] << " is a X bridge" << endl;
		}else
		if(isYBridge[i]){
			cout << seqName[i] << " is a Y bridge" << endl;
		};
	}

		// output subgroups to the bigfile
	cout << "Writing to the big file...";
	cout.flush();
	outputGroups(2, numSeq, isXBridge, isYBridge); // 2 = output stuff to the big file
	cout << "  done." << endl;

	return 0;
}

// end main()
// ---------------------------------------------------------------------

// Read in filenames and cutoff value
void init(ifstream& list_fp, ifstream& matrix_fp, float* cutoff_value){

	string matrixFileName;
	string temp;

	// get list file pointer
	cout << "Enter the LIST file name: ";
	cout.flush();
	cin >> listFileName;
	list_fp.open(listFileName.c_str(), ios::in);
	if(!list_fp){    // if file doesn't exist, quit
		cout << "File " << listFileName << " not found" << endl;
		exit(0);
	}
	
	// get matrix file pointer
	cout << "Enter the MATRIX file name: ";
	cout.flush();
	cin >> matrixFileName;
	matrix_fp.open(matrixFileName.c_str(), ios::in);
	if(!matrix_fp){    // if file doesn't exist, quit
		cout << "File " << matrixFileName << " not found" << endl;
		exit(0);
	}

	// get cutoff value
	cout << "Enter the CUTOFF identity value: ";
	cout.flush();
	cin >> temp;
	*cutoff_value = atof(temp.c_str());

	// get num of seqs
	char ch = 'a';
	maxseq = 1;
	FILE* fp = fopen(listFileName.c_str(), "r");
	while(ch!=EOF){
		ch = fgetc(fp);
		if(ch=='\n') maxseq++;
	}
	fclose(fp);
	seqName.init( maxseq );
	groupUsed.init( maxseq );
	groups.init( maxseq );
	matrix.init( maxseq );
	seqRelation.init( maxseq );

}

// end init()
// ----------------------------------------------------------------------

// put sequence names into seqName array
void getSeqName(ifstream& list_fp, long* numSeq){

	long count = 0;
	while(!list_fp.eof()){
		// if internal array too small, quit
		if(count==maxseq){
			cout << "\n\nTerminated: Internal array too small\n"
			     << "\tThere are too many sequences\n "
			     << "\tPlease input bigger sequence size\n"
			     << "\tSuggested size: (wc -l list_file) + 1"
			     << endl;
			exit(0);
		}
		// read in seq name
		char str[MAXLEN];
		list_fp.getline(str, MAXLEN);    // read seq name
		if(strcmp(str, "")!=0)      // if not null, accept it
			seqName[count++] = str;
	}
	*numSeq = count;
}

// end getSeqName()
// ----------------------------------------------------------------------

void setGroups(ifstream& matrix_fp, float cutoff_value, long numSeq){

	/************************************************************
	 * This part can be explained by this simple example:
	 * 
	 * seqName[0] = 'A'     // A's index is 0
	 * seqName[1] = 'B'     // B's index is 1
	 * ....
	 *
	 * if we have "A B 1" from the matrix file
	 * group_curr = getGroupNum();  // get groupNum from the right
	 *         // group, see more details in the if-else loop below
	 * groupNum[ A's index (0) ] = group_curr;
	 * groupNum[ B's index (1) ] = group_curr;
	 * joined[ A's index (0) ] = true;
	 * joined[ B's index (1) ] = true;
	 * groupUsed[group_curr] = true;
	 *
	 *************************************************************/

	bool joined[numSeq];    // whether a seq is already in a group (seq)
	long groupNum[numSeq];   // if joined, which group? (seq)
	bool groupUsed[numSeq];  // has group being used? empty? (group)
	for(long i=0; i<numSeq; i++){      // init
		joined[i] = false;     // haven't joined
		groupNum[i] = -1;      // no group
		groupUsed[i] = false;  // never been used
	}

	string seq1;    // first seq name
	string seq2;    // second seq name
	float cutoff_curr;    // the cutoff value read from the line
	char unknown[MAXLEN];   // for unknown stuff
	long index1, index2;   // the corresponding indexes from seq1 and seq2
	long group_curr = 0;   // increase this only when creating new group
	long count = 0;        // which line of edges we are currently on

	while(!matrix_fp.eof()){         // while still more relation(s)
		cutoff_curr = 0;
		matrix_fp >> seq1 >> seq2 >> cutoff_curr;  // read in
		matrix_fp.getline(unknown, MAXLEN); 
		count++;
		if(g_verbose){
			cout << count << ") " << seq1 << " "
			     << seq2 << " " << cutoff_curr << endl;
		}
/*
		if(g_verbose){
			for(int i=0; i<maxseq; i++){
				cout << seqName[i] << "\t" << seqRelation[i]
				    << endl;
			}
		}
*/
		if( (count+1)%1000==0 ) cout << "."; cout.flush();
		if(cutoff_curr >= cutoff_value){   // only when qualified
				// get corresponding indexes
			index1 = getSeqIndex(numSeq, seq1);
			index2 = getSeqIndex(numSeq, seq2);
			seqRelation[index1]++;  // assume index1 != index2
			seqRelation[index2]++;
			ostringstream ostr;
			ostr << seq1 << "\t" << seq2 << "\t" << cutoff_curr << endl;
				// store neighbor node info
			seqName.saveNeighbor(index1, index2);
				// this if-else determine what will
				// happen to the seq's and groups
			if(joined[index1]==false && joined[index2]==false){
				// both not in? get to the same new group
				group_curr = getGroupIndex();
				groupNum[index1] = group_curr;
				groups[group_curr].insert(index1);
				groupNum[index2] = group_curr;
				groups[group_curr].insert(index2);
				joined[index1] = true;
				joined[index2] = true;
				matrix[groupNum[index1]].insert(ostr.str());
			}else
			if(joined[index1]==true && joined[index2]==false){
				// join to the existing group
				groupNum[index2] = groupNum[index1];
				joined[index2] = true;
				groups[groupNum[index2]].insert(index2);
				matrix[groupNum[index1]].insert(ostr.str());
			}else
			if(joined[index1]==false && joined[index2]==true){
				// join to the existing group
				groupNum[index1] = groupNum[index2];
				joined[index1] = true;
				groups[groupNum[index1]].insert(index1);
				matrix[groupNum[index2]].insert(ostr.str());
			}else
			if(joined[index1]==true && joined[index2]==true){
			   if(groupNum[index1] != groupNum[index2]){
				// we need to perform a union here
				// by transfering stuff from group2 to group1
				// and then clearing group2
				long groupNum2_old = groupNum[index2];
				set<long>::const_iterator it;
				it=groups[groupNum[index2]].begin();
			        long groupsSize = groups[groupNum[index2]].size();
				for(long i=0;
				    i<groupsSize;
				    i++){
					groups[groupNum[index1]].insert(*it);
					groupNum[*it] = groupNum[index1];
					it++;
				}
				// move relations to the new matrix
				set<string>::const_iterator mit;
				mit=matrix[groupNum2_old].begin();
				long matrixSize = matrix[groupNum2_old].size();
				for(long i=0; i<matrixSize; i++){
					matrix[groupNum[index1]].insert(*mit);
					mit++;
				}
				matrix[groupNum[index1]].insert(ostr.str());
				// clear group2, make it available also
				groups[groupNum2_old].clear();
				groupUsed[groupNum2_old] = false;
			   }else{
				matrix[groupNum[index1]].insert(ostr.str());
			   }
			}
		}
		// if cutoff < cutoff_value,  ignore

		   // for debugging...
		if(g_debug_01){
/*		  for(long i=0; i<numSeq; i++)
			if(joined[i] == true)
				cout << seqName[i] << " joined" << endl;
*/
		  printGroups(numSeq);
		  if(g_debug_02){
		    // to pause the print process
		    char ch;
		    cin.clear();
		    scanf("%c", &ch);
		  }
		  // printGroupsDetails(groupUsed, numSeq);
		}

	}

	// !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	// for those which are alone, assign groups to each of them
	for(long i=0; i<numSeq; i++){
		if(joined[i] == false){
			group_curr = getGroupIndex();
			groupNum[i] = group_curr;
			joined[i] = true;       // for consistency...
			groups[group_curr].insert(i);   // joining...
			   // for debugging...
			if(g_debug_03){
			  printGroupsDetails(numSeq);
			}
		}
	}

}

// end setGroups()
// ----------------------------------------------------------------------

long  getSeqIndex(long numSeq, string seq){

	// sequent search for now...

	for(long i=0; i<numSeq; i++){
		if( seqName[i] == seq ){
//			if(g_verbose) cout << endl;
			return i;
		}
//		if( ((i+1)%1000==0) && g_verbose ) cout << ".";
	}

	cout << "seq " << seq << " doesn't exist..." << endl;
	exit(0);

}

// end getSeqIndex()
// ----------------------------------------------------------------------

void printGroups(long numSeq){

	long count = 0;
	for(long i=0; i<numSeq; i++){
		if(groups[i].size() > 0){
			cout << ++count << ") group #" << i << ": ";
			set<long>::const_iterator it;
			for(it=groups[i].begin(); it!=groups[i].end(); it++){
				cout << seqName[*it] << " ";
			}
			cout << endl;
		}
	}
	cout << endl << endl;

}

// end printGroups()
// ----------------------------------------------------------------------

void printGroupsDetails(long numSeq){

	printGroups(numSeq);

	for(long i=0; i<numSeq; i++){
		if(groupUsed[i] == true)
			cout << "group " << i << " is used" << endl;
	}

	// to pause the print process
	char ch;
	cin.clear();
	scanf("%c", &ch);

}

// end printGroupsDetails()
// ----------------------------------------------------------------------

long getGroupIndex(){

	// sequential search for now...

	// retrieve the next available (empty) group
	for(long i=0; i<maxseq; i++)
		if( groupUsed[i] == false ){
			groupUsed[i] = true;   // assume it will be used
			return i;
		}

	cout << "groups are all filled..." << endl;
	exit(0);

}

// end getGroupIndex()
// ----------------------------------------------------------------------

long outputGroups(int step, long numSeq, bool *isXBridge, bool *isYBridge){

	string subgroup_prefix = "subgroup_";
	string submatrix_prefix = "submatrix_";
	string bigfilename = listFileName + ".group_degree_info";
	ofstream bfp;
	if(step==2){
		bfp.open(bigfilename.c_str(), ios::out);
	}

	long count = 0;
	for(long i=0; i<numSeq; i++){
		if(groups[i].size() > 0){
			// construct the suffix
			string suffix;
			long temp = ++count;
			while(temp!=0){
				string c;
				switch(temp % 10){
					case 0: c = "0"; break;
					case 1: c = "1"; break;
					case 2: c = "2"; break;
					case 3: c = "3"; break;
					case 4: c = "4"; break;
					case 5: c = "5"; break;
					case 6: c = "6"; break;
					case 7: c = "7"; break;
					case 8: c = "8"; break;
					case 9: c = "9"; break;
					default: break;
				}
				temp = temp / 10;  // shift it by one digit
				suffix = c + suffix;
			}
			string filename = subgroup_prefix + suffix;
			ofstream fp;
			if(step==1){
				fp.open(filename.c_str(), ios::out);
			}
			set<long>::const_iterator it;
			for(it=groups[i].begin(); it!=groups[i].end(); it++){
				if(step==1){
					fp << seqName[*it] << endl;
				}
				if(step==2){
					char bridgeSign = '-';
					if(isXBridge[*it]){
						bridgeSign = 'X';
					}
					if(isYBridge[*it]){
						bridgeSign = 'Y';
					}
					if(isXBridge[*it] && isYBridge[*it]){
						bridgeSign = 'Z';
					}
					bfp << seqName[*it] << "\t" << seqRelation[*it]
					    << "\t" << groups[i].size()
					    << "\t" << count
					    << "\t" << bridgeSign;
					if(it==groups[i].begin()){
					  bfp << "\t" << "*****" << endl;
					}else{
					  bfp << endl;
					}
				}
			}
			string mfilename = submatrix_prefix + suffix;
			ofstream mfp;
			if(step==1){
				mfp.open(mfilename.c_str(), ios::out);
			}
			set<string>::const_iterator mit;
			if(step==1){
				for(mit=matrix[i].begin(); mit!=matrix[i].end(); mit++){
					mfp << (*mit);
				}
			}

			if(step==1){
				fp.close();
				mfp.close();
			}
		}
	}

	if(step==2){
		bfp.close();
	}

	return count;

}

// end outputGroups()
// ----------------------------------------------------------------------

void printUsage(){

	cout << "\nUsage: prog [param1 [param2 ........... ] ] \n" << endl;
	cout << "-v	Show infomation of current relations (edges)" << endl;
	cout << "-d1	Show intermediate changes (for edges)" << endl;
	cout << "-d2	pause, add-on for -d1" << endl;
	cout << "-d3	Show intermediate changes (for leftovers)\n" << endl;

	exit(0);

}

// end printUsage()
// ----------------------------------------------------------------------


long checkMatrixGroups(ifstream& matrix_fp, float cutoff_value,
	 long numSeq, string skipSeq)
{

//	seqNameClass seqName;
//	groupUsedClass groupUsed;
	groupsClass groups;
	seqRelationClass seqRelation;
//	seqName.init( maxseq );
//	groupUsed.init( maxseq );
	groups.init( maxseq );
	seqRelation.init( maxseq );

	bool joined[numSeq];    // whether a seq is already in a group (seq)
	long groupNum[numSeq];   // if joined, which group? (seq)
	bool mygroupUsed[numSeq];  // has group being used? empty? (group)
	for(long i=0; i<numSeq; i++){      // init
		joined[i] = false;     // haven't joined
		groupNum[i] = -1;      // no group
		groupUsed[i] = false;  // never been used
	}

	string seq1;    // first seq name
	string seq2;    // second seq name
	float cutoff_curr;    // the cutoff value read from the line
	char unknown[MAXLEN];   // for unknown stuff
	long index1, index2;   // the corresponding indexes from seq1 and seq2
	long group_curr = 0;   // temporary var
	long count = 0;        // which line of edges we are currently on
	long numOfGroups = 0;

	while(!matrix_fp.eof()){         // while still more relation(s)
		cutoff_curr = 0;
		matrix_fp >> seq1 >> seq2 >> cutoff_curr;  // read in
		matrix_fp.getline(unknown, MAXLEN); 
		count++;
		if(seq1==skipSeq || seq2==skipSeq) continue;
		if( (count+1)%1000==0 ) cout << "."; cout.flush();
		if(cutoff_curr >= cutoff_value){   // only when qualified
				// get corresponding indexes
			index1 = getSeqIndex(numSeq, seq1);
			index2 = getSeqIndex(numSeq, seq2);
			seqRelation[index1]++;  // assume index1 != index2
			seqRelation[index2]++;
				// this if-else determine what will
				// happen to the seq's and groups
// cout << count << ") " << seq1 << " " << seq2 << " " << cutoff_curr << "  " << cutoff_value << endl;
			if(joined[index1]==false && joined[index2]==false){
				// both not in? get to the same new group
				group_curr = -1;
				for(long i=0; i<maxseq; i++){
					if( groupUsed[i] == false ){
						groupUsed[i] = true;
						group_curr = i;
						break;
					}
				}
				if(group_curr == -1){
					cout << "groups are all filled..." << endl;
					exit(0);
				}
				groupNum[index1] = group_curr;
				groups[group_curr].insert(index1);
				groupNum[index2] = group_curr;
				groups[group_curr].insert(index2);
				joined[index1] = true;
				joined[index2] = true;
				numOfGroups++;
			}else
			if(joined[index1]==true && joined[index2]==false){
				// join to the existing group
				groupNum[index2] = groupNum[index1];
				joined[index2] = true;
				groups[groupNum[index2]].insert(index2);
			}else
			if(joined[index1]==false && joined[index2]==true){
				// join to the existing group
				groupNum[index1] = groupNum[index2];
				joined[index1] = true;
				groups[groupNum[index1]].insert(index1);
			}else
			if(joined[index1]==true && joined[index2]==true &&
			   groupNum[index1] != groupNum[index2]){
				// we need to perform a union here
				// by transfering stuff from group2 to group1
				// and then clearing group2
				long groupNum2_old = groupNum[index2];
				set<long>::const_iterator it;
				it=groups[groupNum[index2]].begin();
			        long groupsSize = groups[groupNum[index2]].size();
				for(long i=0;
				    i<groupsSize;
				    i++){
					groups[groupNum[index1]].insert(*it);
					groupNum[*it] = groupNum[index1];
					it++;
				}
				// clear group2, make it available also
				groups[groupNum2_old].clear();
				groupUsed[groupNum2_old] = false;
				numOfGroups--;
			}
		}
		// if cutoff < cutoff_value,  ignore

		   // for debugging...
		if(g_debug_01){
/*		  for(long i=0; i<numSeq; i++)
			if(joined[i] == true)
				cout << seqName[i] << " joined" << endl;
*/
		  printGroups(numSeq);
		  if(g_debug_02){
		    // to pause the print process
		    char ch;
		    cin.clear();
		    scanf("%c", &ch);
		  }
		  // printGroupsDetails(groupUsed, numSeq);
		}

	}

	return numOfGroups;

}

// end checkMatrixGroups
// ----------------------------------------------------------------------



